@@ -10,7 +10,6 @@ Operating system specific functionality.
 from __future__ import absolute_import
 import socket
 
-import scapy.consts
 from scapy.consts import LINUX, SOLARIS, WINDOWS, BSD
 from scapy.error import Scapy_Exception
 from scapy.config import conf, _set_conf_sockets
@@ -24,7 +23,7 @@ def str2mac(s):
 
 
 if not WINDOWS:
-    if not conf.use_pcap and not conf.use_dnet:
+    if not conf.use_pcap:
         from scapy.arch.bpf.core import get_if_raw_addr
 
 
@@ -56,18 +55,19 @@ if LINUX:
     from scapy.arch.linux import *  # noqa F403
 elif BSD:
     from scapy.arch.unix import read_routes, read_routes6, in6_getifaddr  # noqa: F401, E501
-
-    if not conf.use_pcap or conf.use_dnet:
-        from scapy.arch.bpf.core import *  # noqa F403
+    from scapy.arch.bpf.core import *  # noqa F403
+    if not conf.use_pcap:
+        # Native
         from scapy.arch.bpf.supersocket import * # noqa F403
         conf.use_bpf = True
 elif SOLARIS:
     from scapy.arch.solaris import *  # noqa F403
 elif WINDOWS:
     from scapy.arch.windows import *  # noqa F403
+    from scapy.arch.windows.native import *  # noqa F403
 
 if conf.iface is None:
-    conf.iface = scapy.consts.LOOPBACK_INTERFACE
+    conf.iface = conf.loopback_name
 
 _set_conf_sockets()  # Apply config
 
