@@ -5,22 +5,23 @@ Scapy *BSD native support - core
 """
 
 from __future__ import absolute_import
-from scapy.config import conf
-from scapy.error import Scapy_Exception, warning
-from scapy.data import ARPHDR_LOOPBACK, ARPHDR_ETHER
-from scapy.arch.common import get_if, compile_filter
-from scapy.consts import LOOPBACK_NAME
-
-from scapy.arch.bpf.consts import BIOCSETF, SIOCGIFFLAGS, BIOCSETIF
-
-import os
-import socket
-import fcntl
-import struct
 
 from ctypes import cdll, cast, pointer
 from ctypes import c_int, c_ulong, c_char_p
 from ctypes.util import find_library
+import fcntl
+import os
+import re
+import socket
+import struct
+import subprocess
+
+from scapy.arch.bpf.consts import BIOCSETF, SIOCGIFFLAGS, BIOCSETIF
+from scapy.arch.common import get_if, compile_filter
+from scapy.compat import plain_str
+from scapy.config import conf
+from scapy.data import ARPHDR_LOOPBACK, ARPHDR_ETHER
+from scapy.error import Scapy_Exception, warning
 from scapy.modules.six.moves import range
 
 
@@ -37,20 +38,29 @@ def get_if_raw_addr(ifname):
     """Returns the IPv4 address configured on 'ifname', packed with inet_pton."""  # noqa: E501
 
     # Get ifconfig output
-    try:
-        fd = os.popen("%s %s" % (conf.prog.ifconfig, ifname))
-    except OSError as msg:
-        warning("Failed to execute ifconfig: (%s)", msg)
+    subproc = subprocess.Popen(
+        [conf.prog.ifconfig, ifname],
+        close_fds=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
+    )
+    stdout, stderr = subproc.communicate()
+    if subproc.returncode:
+        warning("Failed to execute ifconfig: (%s)", plain_str(stderr))
         return b"\0\0\0\0"
-
     # Get IPv4 addresses
-    addresses = [l for l in fd if l.find("netmask") >= 0]
+
+    addresses = [
+        line for line in plain_str(stdout).splitlines()
+        if "inet " in line
+    ]
+
     if not addresses:
         warning("No IPv4 address found on %s !", ifname)
         return b"\0\0\0\0"
 
     # Pack the first address
     address = addresses[0].split(' ')[1]
+    if '/' in address:  # NetBSD 8.0
+        address = address.split("/")[0]
     return socket.inet_pton(socket.AF_INET, address)
 
 
@@ -60,19 +70,25 @@ def get_if_raw_hwaddr(ifname):
     NULL_MAC_ADDRESS = b'\x00' * 6
 
     # Handle the loopback interface separately
-    if ifname == LOOPBACK_NAME:
+    if ifname == conf.loopback_name:
         return (ARPHDR_LOOPBACK, NULL_MAC_ADDRESS)
 
     # Get ifconfig output
-    try:
-        fd = os.popen("%s %s" % (conf.prog.ifconfig, ifname))
-    except OSError as msg:
-        raise Scapy_Exception("Failed to execute ifconfig: (%s)" % msg)
+    subproc = subprocess.Popen(
+        [conf.prog.ifconfig, ifname],
+        close_fds=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
+    )
+    stdout, stderr = subproc.communicate()
+    if subproc.returncode:
+        raise Scapy_Exception("Failed to execute ifconfig: (%s)" %
+                              (plain_str(stderr)))
 
     # Get MAC addresses
-    addresses = [l for l in fd.readlines() if l.find("ether") >= 0 or
-                 l.find("lladdr") >= 0 or
-                 l.find("address") >= 0]
+    addresses = [
+        line for line in plain_str(stdout).splitlines() if (
+            "ether" in line or "lladdr" in line or "address" in line
+        )
+    ]
     if not addresses:
         raise Scapy_Exception("No MAC address found on %s !" % ifname)
 
@@ -113,17 +129,25 @@ def get_if_list():
     """Returns a list containing all network interfaces."""
 
     # Get ifconfig output
-    try:
-        fd = os.popen("%s -a" % conf.prog.ifconfig)
-    except OSError as msg:
-        raise Scapy_Exception("Failed to execute ifconfig: (%s)" % msg)
+    subproc = subprocess.Popen(
+        [conf.prog.ifconfig],
+        close_fds=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
+    )
+    stdout, stderr = subproc.communicate()
+    if subproc.returncode:
+        raise Scapy_Exception("Failed to execute ifconfig: (%s)" %
+                              (plain_str(stderr)))
 
-    # Get interfaces
-    interfaces = [line[:line.find(':')] for line in fd.readlines()
-                  if ": flags" in line.lower()]
+    interfaces = [
+        line[:line.find(':')] for line in plain_str(stdout).splitlines()
+        if ": flags" in line.lower()
+    ]
     return interfaces
 
 
+_IFNUM = re.compile(r"([0-9]*)([ab]?)$")
+
+
 def get_working_ifaces():
     """
     Returns an ordered list of interfaces that could be used with BPF.
@@ -139,7 +163,7 @@ def get_working_ifaces():
     for ifname in get_if_list():
 
         # Unlike pcap_findalldevs(), we do not care of loopback interfaces.
-        if ifname == LOOPBACK_NAME:
+        if ifname == conf.loopback_name:
             continue
 
         # Get interface flags
@@ -154,24 +178,27 @@ def get_working_ifaces():
         if ifflags & 0x1:  # IFF_UP
 
             # Get a BPF handle
-            fd, _ = get_dev_bpf()
+            fd = get_dev_bpf()[0]
             if fd is None:
                 raise Scapy_Exception("No /dev/bpf are available !")
 
             # Check if the interface can be used
             try:
-                fcntl.ioctl(fd, BIOCSETIF, struct.pack("16s16x", ifname.encode()))  # noqa: E501
-                interfaces.append((ifname, int(ifname[-1])))
+                fcntl.ioctl(fd, BIOCSETIF, struct.pack("16s16x",
+                                                       ifname.encode()))
             except IOError:
                 pass
-
-            # Close the file descriptor
-            os.close(fd)
+            else:
+                ifnum, ifab = _IFNUM.search(ifname).groups()
+                interfaces.append((ifname, int(ifnum) if ifnum else -1, ifab))
+            finally:
+                # Close the file descriptor
+                os.close(fd)
 
     # Sort to mimic pcap_findalldevs() order
-    interfaces.sort(key=lambda elt: elt[1])
+    interfaces.sort(key=lambda elt: (elt[1], elt[2], elt[0]))
 
-    return interfaces
+    return [iface[0] for iface in interfaces]
 
 
 def get_working_if():
@@ -180,5 +207,5 @@ def get_working_if():
     ifaces = get_working_ifaces()
     if not ifaces:
         # A better interface will be selected later using the routing table
-        return LOOPBACK_NAME
-    return ifaces[0][0]
+        return conf.loopback_name
+    return ifaces[0]
