@@ -1,5 +1,3 @@
-#! /usr/bin/env python
-
 # Copyright (C) 2018 Leonardo Monteiro <decastromonteiro@gmail.com>
 #               2017 Alexis Sultan    <alexis.sultan@sfr.com>
 #               2017 Alessio Deiana <adeiana@gmail.com>
@@ -15,7 +13,8 @@ from __future__ import absolute_import
 import struct
 
 
-from scapy.compat import chb, orb
+from scapy.compat import chb, orb, bytes_encode
+from scapy.config import conf
 from scapy.error import warning
 from scapy.fields import BitEnumField, BitField, ByteEnumField, ByteField, \
     ConditionalField, FieldLenField, FieldListField, FlagsField, IntField, \
@@ -148,6 +147,7 @@ ExtensionHeadersTypes = {
     1: "Reserved",
     2: "Reserved",
     64: "UDP Port",
+    133: "PDU Session Container",
     192: "PDCP PDU Number",
     193: "Reserved",
     194: "Reserved"
@@ -168,18 +168,22 @@ class TBCDByteField(StrFixedLenField):
             if left == 0xf:
                 ret.append(TBCD_TO_ASCII[right:right + 1])
             else:
-                ret += [TBCD_TO_ASCII[right:right + 1], TBCD_TO_ASCII[left:left + 1]]  # noqa: E501
+                ret += [
+                    TBCD_TO_ASCII[right:right + 1],
+                    TBCD_TO_ASCII[left:left + 1]
+                ]
         return b"".join(ret)
 
     def i2m(self, pkt, val):
-        val = str(val)
-        ret_string = ""
+        if not isinstance(val, bytes):
+            val = bytes_encode(val)
+        ret_string = b""
         for i in range(0, len(val), 2):
             tmp = val[i:i + 2]
             if len(tmp) == 2:
-                ret_string += chr(int(tmp[1] + tmp[0], 16))
+                ret_string += chb(int(tmp[::-1], 16))
             else:
-                ret_string += chr(int("F" + tmp[0], 16))
+                ret_string += chb(int(b"F" + tmp[:1], 16))
         return ret_string
 
 
@@ -218,9 +222,16 @@ class GTPHeader(Packet):
                    ByteEnumField("gtp_type", None, GTPmessageType),
                    ShortField("length", None),
                    IntField("teid", 0),
-                   ConditionalField(XBitField("seq", 0, 16), lambda pkt:pkt.E == 1 or pkt.S == 1 or pkt.PN == 1),  # noqa: E501
-                   ConditionalField(ByteField("npdu", 0), lambda pkt:pkt.E == 1 or pkt.S == 1 or pkt.PN == 1),  # noqa: E501
-                   ConditionalField(ByteEnumField("next_ex", 0, ExtensionHeadersTypes), lambda pkt:pkt.E == 1 or pkt.S == 1 or pkt.PN == 1), ]  # noqa: E501
+                   ConditionalField(
+                       XBitField("seq", 0, 16),
+                       lambda pkt:pkt.E == 1 or pkt.S == 1 or pkt.PN == 1),
+                   ConditionalField(
+                       ByteField("npdu", 0),
+                       lambda pkt:pkt.E == 1 or pkt.S == 1 or pkt.PN == 1),
+                   ConditionalField(
+                       ByteEnumField("next_ex", 0, ExtensionHeadersTypes),
+                       lambda pkt:pkt.E == 1 or pkt.S == 1 or pkt.PN == 1),
+                   ]
 
     def post_build(self, p, pay):
         p += pay
@@ -230,11 +241,15 @@ class GTPHeader(Packet):
         return p
 
     def hashret(self):
-        return struct.pack("B", self.version) + self.payload.hashret()
+        hsh = struct.pack("B", self.version)
+        if self.seq:
+            hsh += struct.pack("H", self.seq)
+        return hsh + self.payload.hashret()
 
     def answers(self, other):
         return (isinstance(other, GTPHeader) and
                 self.version == other.version and
+                (not self.seq or self.seq == other.seq) and
                 self.payload.answers(other.payload))
 
     @classmethod
@@ -259,6 +274,11 @@ class GTP_U_Header(GTPHeader):
     def guess_payload_class(self, payload):
         # Snooped from Wireshark
         # https://github.com/boundary/wireshark/blob/07eade8124fd1d5386161591b52e177ee6ea849f/epan/dissectors/packet-gtp.c#L8195  # noqa: E501
+        if self.E == 1:
+            if self.next_ex == 0x85:
+                return GTPPDUSessionContainer
+            return GTPHeader.guess_payload_class(self, payload)
+
         if self.gtp_type == 255:
             sub_proto = orb(payload[0])
             if sub_proto >= 0x45 and sub_proto <= 0x4e:
@@ -285,19 +305,71 @@ GTPforcedTypes = {
 }
 
 
+class GTPPDUSessionContainer(Packet):
+    name = "GTP PDU Session Container"
+    fields_desc = [ByteField("ExtHdrLen", None),
+                   BitField("type", 0, 4),
+                   BitField("spare1", 0, 4),
+                   BitField("P", 0, 1),
+                   BitField("R", 0, 1),
+                   BitField("QFI", 0, 6),
+                   ConditionalField(XBitField("PPI", 0, 3),
+                                    lambda pkt: pkt.P == 1),
+                   ConditionalField(XBitField("spare2", 0, 5),
+                                    lambda pkt: pkt.P == 1),
+                   ConditionalField(ByteField("pad1", 0),
+                                    lambda pkt: pkt.P == 1),
+                   ConditionalField(ByteField("pad2", 0),
+                                    lambda pkt: pkt.P == 1),
+                   ConditionalField(ByteField("pad3", 0),
+                                    lambda pkt: pkt.P == 1),
+                   ConditionalField(StrLenField(
+                       "extraPadding",
+                       "",
+                       length_from=lambda pkt: 4 * (pkt.ExtHdrLen) - 4),
+                       lambda pkt: pkt.ExtHdrLen and pkt.ExtHdrLen > 1),
+                   ByteEnumField("NextExtHdr", 0, ExtensionHeadersTypes), ]
+
+    def guess_payload_class(self, payload):
+        if self.NextExtHdr == 0:
+            sub_proto = orb(payload[0])
+            if sub_proto >= 0x45 and sub_proto <= 0x4e:
+                return IP
+            elif (sub_proto & 0xf0) == 0x60:
+                return IPv6
+            else:
+                return PPP
+        return GTPHeader.guess_payload_class(self, payload)
+
+    def post_build(self, p, pay):
+        p += pay
+        if self.ExtHdrLen is None:
+            if self.P == 1:
+                hdr_len = 2
+            else:
+                hdr_len = 1
+            p = struct.pack("!B", hdr_len) + p[1:]
+        return p
+
+
 class GTPEchoRequest(Packet):
     # 3GPP TS 29.060 V9.1.0 (2009-12)
     name = "GTP Echo Request"
 
-    def hashret(self):
-        return struct.pack("H", self.seq)
-
 
 class IE_Base(Packet):
-
     def extract_padding(self, pkt):
         return "", pkt
 
+    def post_build(self, p, pay):
+        if self.fields_desc[1].name == "length":
+            if self.length is None:
+                tmp_len = len(p)
+                if isinstance(self.payload, conf.padding_layer):
+                    tmp_len += len(self.payload.load)
+                p = p[:1] + struct.pack("!H", tmp_len - 2) + p[3:]
+        return p + pay
+
 
 class IE_Cause(IE_Base):
     name = "Cause"
@@ -446,7 +518,9 @@ class APNStrLenField(StrLenField):
         return s
 
     def i2m(self, pkt, s):
-        s = b"".join(chb(len(x)) + x for x in s.split("."))
+        if not isinstance(s, bytes):
+            s = bytes_encode(s)
+        s = b"".join(chb(len(x)) + x for x in s.split(b"."))
         return s
 
 
@@ -475,8 +549,17 @@ class IE_ProtocolConfigurationOptions(IE_Base):
 class IE_GSNAddress(IE_Base):
     name = "GSN Address"
     fields_desc = [ByteEnumField("ietype", 133, IEType),
-                   ShortField("length", 4),
-                   IPField("address", RandIP())]
+                   ShortField("length", None),
+                   ConditionalField(IPField("ipv4_address", RandIP()),
+                                    lambda pkt: pkt.length == 4),
+                   ConditionalField(IP6Field("ipv6_address", '::1'),
+                                    lambda pkt: pkt.length == 16)]
+
+    def post_build(self, p, pay):
+        if self.length is None:
+            tmp_len = len(p) - 3
+            p = p[:2] + struct.pack("!B", tmp_len) + p[3:]
+        return p
 
 
 class IE_MSInternationalNumber(IE_Base):
@@ -520,82 +603,82 @@ class IE_QoS(IE_Base):
                    ByteField("allocation_retention_prioiry", 1),
 
                    ConditionalField(XBitField("spare", 0x00, 2),
-                                    lambda pkt: pkt.length > 1),
+                                    lambda p: p.length and p.length > 1),
                    ConditionalField(XBitField("delay_class", 0x000, 3),
-                                    lambda pkt: pkt.length > 1),
+                                    lambda p: p.length and p.length > 1),
                    ConditionalField(XBitField("reliability_class", 0x000, 3),
-                                    lambda pkt: pkt.length > 1),
+                                    lambda p: p.length and p.length > 1),
 
                    ConditionalField(XBitField("peak_troughput", 0x0000, 4),
-                                    lambda pkt: pkt.length > 2),
+                                    lambda p: p.length and p.length > 2),
                    ConditionalField(BitField("spare", 0, 1),
-                                    lambda pkt: pkt.length > 2),
+                                    lambda p: p.length and p.length > 2),
                    ConditionalField(XBitField("precedence_class", 0x000, 3),
-                                    lambda pkt: pkt.length > 2),
+                                    lambda p: p.length and p.length > 2),
 
                    ConditionalField(XBitField("spare", 0x000, 3),
-                                    lambda pkt: pkt.length > 3),
+                                    lambda p: p.length and p.length > 3),
                    ConditionalField(XBitField("mean_troughput", 0x00000, 5),
-                                    lambda pkt: pkt.length > 3),
+                                    lambda p: p.length and p.length > 3),
 
                    ConditionalField(XBitField("traffic_class", 0x000, 3),
-                                    lambda pkt: pkt.length > 4),
+                                    lambda p: p.length and p.length > 4),
                    ConditionalField(XBitField("delivery_order", 0x00, 2),
-                                    lambda pkt: pkt.length > 4),
+                                    lambda p: p.length and p.length > 4),
                    ConditionalField(XBitField("delivery_of_err_sdu", 0x000, 3),
-                                    lambda pkt: pkt.length > 4),
+                                    lambda p: p.length and p.length > 4),
 
                    ConditionalField(ByteField("max_sdu_size", None),
-                                    lambda pkt: pkt.length > 5),
+                                    lambda p: p.length and p.length > 5),
                    ConditionalField(ByteField("max_bitrate_up", None),
-                                    lambda pkt: pkt.length > 6),
+                                    lambda p: p.length and p.length > 6),
                    ConditionalField(ByteField("max_bitrate_down", None),
-                                    lambda pkt: pkt.length > 7),
+                                    lambda p: p.length and p.length > 7),
 
                    ConditionalField(XBitField("redidual_ber", 0x0000, 4),
-                                    lambda pkt: pkt.length > 8),
+                                    lambda p: p.length and p.length > 8),
                    ConditionalField(XBitField("sdu_err_ratio", 0x0000, 4),
-                                    lambda pkt: pkt.length > 8),
+                                    lambda p: p.length and p.length > 8),
                    ConditionalField(XBitField("transfer_delay", 0x00000, 6),
-                                    lambda pkt: pkt.length > 9),
+                                    lambda p: p.length and p.length > 9),
                    ConditionalField(XBitField("traffic_handling_prio",
                                               0x000,
                                               2),
-                                    lambda pkt: pkt.length > 9),
+                                    lambda p: p.length and p.length > 9),
 
                    ConditionalField(ByteField("guaranteed_bit_rate_up", None),
-                                    lambda pkt: pkt.length > 10),
+                                    lambda p: p.length and p.length > 10),
                    ConditionalField(ByteField("guaranteed_bit_rate_down",
                                               None),
-                                    lambda pkt: pkt.length > 11),
+                                    lambda p: p.length and p.length > 11),
 
                    ConditionalField(XBitField("spare", 0x000, 3),
-                                    lambda pkt: pkt.length > 12),
+                                    lambda p: p.length and p.length > 12),
                    ConditionalField(BitField("signaling_indication", 0, 1),
-                                    lambda pkt: pkt.length > 12),
+                                    lambda p: p.length and p.length > 12),
                    ConditionalField(XBitField("source_stats_desc", 0x0000, 4),
-                                    lambda pkt: pkt.length > 12),
+                                    lambda p: p.length and p.length > 12),
 
                    ConditionalField(ByteField("max_bitrate_down_ext", None),
-                                    lambda pkt: pkt.length > 13),
+                                    lambda p: p.length and p.length > 13),
                    ConditionalField(ByteField("guaranteed_bitrate_down_ext",
                                               None),
-                                    lambda pkt: pkt.length > 14),
+                                    lambda p: p.length and p.length > 14),
                    ConditionalField(ByteField("max_bitrate_up_ext", None),
-                                    lambda pkt: pkt.length > 15),
+                                    lambda p: p.length and p.length > 15),
                    ConditionalField(ByteField("guaranteed_bitrate_up_ext",
                                               None),
-                                    lambda pkt: pkt.length > 16),
+                                    lambda p: p.length and p.length > 16),
                    ConditionalField(ByteField("max_bitrate_down_ext2", None),
-                                    lambda pkt: pkt.length > 17),
+                                    lambda p: p.length and p.length > 17),
                    ConditionalField(ByteField("guaranteed_bitrate_down_ext2",
                                               None),
-                                    lambda pkt: pkt.length > 18),
+                                    lambda p: p.length and p.length > 18),
                    ConditionalField(ByteField("max_bitrate_up_ext2", None),
-                                    lambda pkt: pkt.length > 19),
+                                    lambda p: p.length and p.length > 19),
                    ConditionalField(ByteField("guaranteed_bitrate_up_ext2",
                                               None),
-                                    lambda pkt: pkt.length > 20)]
+                                    lambda p: p.length and p.length > 20)]
 
 
 class IE_CommonFlags(IE_Base):
@@ -790,35 +873,26 @@ class GTPEchoResponse(Packet):
     name = "GTP Echo Response"
     fields_desc = [PacketListField("IE_list", [], IE_Dispatcher)]
 
-    def hashret(self):
-        return struct.pack("H", self.seq)
-
     def answers(self, other):
-        return self.seq == other.seq
+        return isinstance(other, GTPEchoRequest)
 
 
 class GTPCreatePDPContextRequest(Packet):
     # 3GPP TS 29.060 V9.1.0 (2009-12)
     name = "GTP Create PDP Context Request"
-    fields_desc = [PacketListField("IE_list", [IE_TEIDI(), IE_NSAPI(), IE_GSNAddress(),  # noqa: E501
-                                               IE_GSNAddress(),
+    fields_desc = [PacketListField("IE_list", [IE_TEIDI(), IE_NSAPI(), IE_GSNAddress(length=4, ipv4_address=RandIP()),  # noqa: E501
+                                               IE_GSNAddress(length=4, ipv4_address=RandIP()),  # noqa: E501
                                                IE_NotImplementedTLV(ietype=135, length=15, data=RandString(15))],  # noqa: E501
                                    IE_Dispatcher)]
 
-    def hashret(self):
-        return struct.pack("H", self.seq)
-
 
 class GTPCreatePDPContextResponse(Packet):
     # 3GPP TS 29.060 V9.1.0 (2009-12)
     name = "GTP Create PDP Context Response"
     fields_desc = [PacketListField("IE_list", [], IE_Dispatcher)]
 
-    def hashret(self):
-        return struct.pack("H", self.seq)
-
     def answers(self, other):
-        return self.seq == other.seq
+        return isinstance(other, GTPCreatePDPContextRequest)
 
 
 class GTPUpdatePDPContextRequest(Packet):
@@ -846,17 +920,14 @@ class GTPUpdatePDPContextRequest(Packet):
         IE_PrivateExtension()],
         IE_Dispatcher)]
 
-    def hashret(self):
-        return struct.pack("H", self.seq)
-
 
 class GTPUpdatePDPContextResponse(Packet):
     # 3GPP TS 29.060 V9.1.0 (2009-12)
     name = "GTP Update PDP Context Response"
     fields_desc = [PacketListField("IE_list", None, IE_Dispatcher)]
 
-    def hashret(self):
-        return struct.pack("H", self.seq)
+    def answers(self, other):
+        return isinstance(other, GTPUpdatePDPContextRequest)
 
 
 class GTPErrorIndication(Packet):
@@ -884,7 +955,7 @@ class GTPPDUNotificationRequest(Packet):
                                                IE_TEICP(TEICI=RandInt()),
                                                IE_EndUserAddress(PDPTypeNumber=0x21),  # noqa: E501
                                                IE_AccessPointName(),
-                                               IE_GSNAddress(address="127.0.0.1"),  # noqa: E501
+                                               IE_GSNAddress(ipv4_address="127.0.0.1"),  # noqa: E501
                                                ], IE_Dispatcher)]
 
 
@@ -923,6 +994,8 @@ bind_bottom_up(UDP, GTP_U_Header, dport=2152)
 bind_bottom_up(UDP, GTP_U_Header, sport=2152)
 bind_layers(UDP, GTP_U_Header, dport=2152, sport=2152)
 bind_layers(GTP_U_Header, GTPErrorIndication, gtp_type=26, S=1)
+bind_layers(GTP_U_Header, GTPPDUSessionContainer,
+            gtp_type=255, E=1, next_ex=0x85)
 bind_top_down(GTP_U_Header, IP, gtp_type=255)
 bind_top_down(GTP_U_Header, IPv6, gtp_type=255)
 bind_top_down(GTP_U_Header, PPP, gtp_type=255)
