@@ -41,13 +41,13 @@
 
         Currently it (partially) supports the packaging/encoding
         requirements of the following RFCs:
-         * RFC 1195 (only the TCP/IP related part)
-         * RFC 3358 (optional checksums)
-         * RFC 5301 (dynamic hostname extension)
-         * RFC 5302 (domain-wide prefix distribution)
-         * RFC 5303 (three-way handshake)
-         * RFC 5304 (cryptographic authentication)
-         * RFC 5308 (routing IPv6 with IS-IS)
+        * RFC 1195 (only the TCP/IP related part)
+        * RFC 3358 (optional checksums)
+        * RFC 5301 (dynamic hostname extension)
+        * RFC 5302 (domain-wide prefix distribution)
+        * RFC 5303 (three-way handshake)
+        * RFC 5304 (cryptographic authentication)
+        * RFC 5308 (routing IPv6 with IS-IS)
 
     :TODO:
 
@@ -80,8 +80,6 @@ from scapy.compat import orb, hex_bytes
 
 EXT_VERSION = "v0.0.2"
 
-conf.debug_dissector = True
-
 
 #######################################################################
 #   ISIS Utilities + Fields                                           #
@@ -150,6 +148,7 @@ class _ISIS_IdFieldBase(Field):
 
 class _ISIS_RandId(RandString):
     def __init__(self, template):
+        RandString.__init__(self)
         self.bytecount = template.count("*")
         self.format = template.replace("*", "%02X")
 
@@ -167,8 +166,16 @@ class _ISIS_RandId(RandString):
 
 class _ISIS_RandAreaId(_ISIS_RandId):
     def __init__(self, bytecount=None):
-        self.bytecount = random.randint(1, 13) if bytecount is None else bytecount  # noqa: E501
-        self.format = "%02X" + (".%02X%02X" * ((self.bytecount - 1) // 2)) + ("" if ((self.bytecount - 1) % 2) == 0 else ".%02X")  # noqa: E501
+        template = "*" + (
+            ".**" * ((self.bytecount - 1) // 2)
+        ) + (
+            "" if ((self.bytecount - 1) % 2) == 0 else ".*"
+        )
+        super(_ISIS_RandAreaId, self).__init__(template)
+        if bytecount is None:
+            self.bytecount = random.randint(1, 13)
+        else:
+            self.bytecount = bytecount
 
 
 class ISIS_AreaIdField(Field):
