@@ -9,6 +9,7 @@ SuperSocket.
 
 from __future__ import absolute_import
 from select import select, error as select_error
+import ctypes
 import errno
 import os
 import socket
@@ -17,14 +18,16 @@ import time
 
 from scapy.config import conf
 from scapy.consts import LINUX, DARWIN, WINDOWS
-from scapy.data import MTU, ETH_P_IP
-from scapy.compat import raw
+from scapy.data import MTU, ETH_P_IP, SOL_PACKET, SO_TIMESTAMPNS
+from scapy.compat import raw, bytes_encode
 from scapy.error import warning, log_runtime
 import scapy.modules.six as six
 import scapy.packet
 from scapy.utils import PcapReader, tcpdump
 
 
+# Utils
+
 class _SuperSocket_metaclass(type):
     def __repr__(self):
         if self.desc is not None:
@@ -33,10 +36,31 @@ class _SuperSocket_metaclass(type):
             return "<%s>" % self.__name__
 
 
+# Used to get ancillary data
+PACKET_AUXDATA = 8
+ETH_P_8021Q = 0x8100
+TP_STATUS_VLAN_VALID = 1 << 4
+
+
+class tpacket_auxdata(ctypes.Structure):
+    _fields_ = [
+        ("tp_status", ctypes.c_uint),
+        ("tp_len", ctypes.c_uint),
+        ("tp_snaplen", ctypes.c_uint),
+        ("tp_mac", ctypes.c_ushort),
+        ("tp_net", ctypes.c_ushort),
+        ("tp_vlan_tci", ctypes.c_ushort),
+        ("tp_padding", ctypes.c_ushort),
+    ]
+
+
+# SuperSocket
+
 class SuperSocket(six.with_metaclass(_SuperSocket_metaclass)):
     desc = None
     closed = 0
-    read_allowed_exceptions = ()
+    nonblocking_socket = False
+    auxdata_available = False
 
     def __init__(self, family=socket.AF_INET, type=socket.SOCK_STREAM, proto=0):  # noqa: E501
         self.ins = socket.socket(family, type, proto)
@@ -45,10 +69,64 @@ class SuperSocket(six.with_metaclass(_SuperSocket_metaclass)):
 
     def send(self, x):
         sx = raw(x)
-        if hasattr(x, "sent_time"):
+        try:
             x.sent_time = time.time()
+        except AttributeError:
+            pass
         return self.outs.send(sx)
 
+    if six.PY2:
+        def _recv_raw(self, sock, x):
+            """Internal function to receive a Packet"""
+            pkt, sa_ll = sock.recvfrom(x)
+            return pkt, sa_ll, None
+    else:
+        def _recv_raw(self, sock, x):
+            """Internal function to receive a Packet,
+            and process ancillary data.
+            """
+            timestamp = None
+            if not self.auxdata_available:
+                pkt, _, _, sa_ll = sock.recvmsg(x)
+                return pkt, sa_ll, timestamp
+            flags_len = socket.CMSG_LEN(4096)
+            pkt, ancdata, flags, sa_ll = sock.recvmsg(x, flags_len)
+            if not pkt:
+                return pkt, sa_ll, timestamp
+            for cmsg_lvl, cmsg_type, cmsg_data in ancdata:
+                # Check available ancillary data
+                if (cmsg_lvl == SOL_PACKET and cmsg_type == PACKET_AUXDATA):
+                    # Parse AUXDATA
+                    try:
+                        auxdata = tpacket_auxdata.from_buffer_copy(cmsg_data)
+                    except ValueError:
+                        # Note: according to Python documentation, recvmsg()
+                        #       can return a truncated message. A ValueError
+                        #       exception likely indicates that Auxiliary
+                        #       Data is not supported by the Linux kernel.
+                        return pkt, sa_ll, timestamp
+                    if auxdata.tp_vlan_tci != 0 or \
+                            auxdata.tp_status & TP_STATUS_VLAN_VALID:
+                        # Insert VLAN tag
+                        tag = struct.pack(
+                            "!HH",
+                            ETH_P_8021Q,
+                            auxdata.tp_vlan_tci
+                        )
+                        pkt = pkt[:12] + tag + pkt[12:]
+                elif cmsg_lvl == socket.SOL_SOCKET and \
+                        cmsg_type == SO_TIMESTAMPNS:
+                    length = len(cmsg_data)
+                    if length == 16:  # __kernel_timespec
+                        tmp = struct.unpack("ll", cmsg_data)
+                    elif length == 8:  # timespec
+                        tmp = struct.unpack("ii", cmsg_data)
+                    else:
+                        log_runtime.warning("Unknown timespec format.. ?!")
+                        continue
+                    timestamp = tmp[0] + tmp[1] * 1e-9
+            return pkt, sa_ll, timestamp
+
     def recv_raw(self, x=MTU):
         """Returns a tuple containing (cls, pkt_data, time)"""
         return conf.raw_layer, self.ins.recv(x), None
@@ -63,9 +141,12 @@ class SuperSocket(six.with_metaclass(_SuperSocket_metaclass)):
             raise
         except Exception:
             if conf.debug_dissector:
+                from scapy.sendrecv import debug
+                debug.crashed_on = (cls, val)
                 raise
             pkt = conf.raw_layer(val)
-        pkt.time = ts
+        if ts:
+            pkt.time = ts
         return pkt
 
     def fileno(self):
@@ -75,12 +156,12 @@ class SuperSocket(six.with_metaclass(_SuperSocket_metaclass)):
         if self.closed:
             return
         self.closed = True
-        if hasattr(self, "outs"):
-            if not hasattr(self, "ins") or self.ins != self.outs:
-                if self.outs and (WINDOWS or self.outs.fileno() != -1):
+        if getattr(self, "outs", None):
+            if getattr(self, "ins", None) != self.outs:
+                if WINDOWS or self.outs.fileno() != -1:
                     self.outs.close()
-        if hasattr(self, "ins"):
-            if self.ins and (WINDOWS or self.ins.fileno() != -1):
+        if getattr(self, "ins", None):
+            if WINDOWS or self.ins.fileno() != -1:
                 self.ins.close()
 
     def sr(self, *args, **kargs):
@@ -99,22 +180,24 @@ class SuperSocket(six.with_metaclass(_SuperSocket_metaclass)):
         from scapy import sendrecv
         return sendrecv.sniff(opened_socket=self, *args, **kargs)
 
+    def tshark(self, *args, **kargs):
+        from scapy import sendrecv
+        return sendrecv.tshark(opened_socket=self, *args, **kargs)
+
     @staticmethod
     def select(sockets, remain=conf.recv_poll_rate):
         """This function is called during sendrecv() routine to select
         the available sockets.
 
-        params:
-         - sockets: an array of sockets that need to be selected
-        returns:
-         - an array of sockets that were selected
-         - the function to be called next to get the packets (i.g. recv)
+        :param sockets: an array of sockets that need to be selected
+        :returns: an array of sockets that were selected and
+            the function to be called next to get the packets (i.g. recv)
         """
         try:
             inp, _, _ = select(sockets, [], [], remain)
         except (IOError, select_error) as exc:
             # select.error has no .errno attribute
-            if exc.args[0] != errno.EINTR:
+            if not exc.args or exc.args[0] != errno.EINTR:
                 raise
         return inp, None
 
@@ -137,11 +220,27 @@ class L3RawSocket(SuperSocket):
         self.outs = socket.socket(socket.AF_INET, socket.SOCK_RAW, socket.IPPROTO_RAW)  # noqa: E501
         self.outs.setsockopt(socket.SOL_IP, socket.IP_HDRINCL, 1)
         self.ins = socket.socket(socket.AF_PACKET, socket.SOCK_RAW, socket.htons(type))  # noqa: E501
+        self.iface = iface
         if iface is not None:
-            self.ins.bind((iface, type))
+            self.ins.bind((self.iface, type))
+        if not six.PY2:
+            try:
+                # Receive Auxiliary Data (VLAN tags)
+                self.ins.setsockopt(SOL_PACKET, PACKET_AUXDATA, 1)
+                self.ins.setsockopt(
+                    socket.SOL_SOCKET,
+                    SO_TIMESTAMPNS,
+                    1
+                )
+                self.auxdata_available = True
+            except OSError:
+                # Note: Auxiliary Data is only supported since
+                #       Linux 2.6.21
+                msg = "Your Linux Kernel does not support Auxiliary Data!"
+                log_runtime.info(msg)
 
     def recv(self, x=MTU):
-        pkt, sa_ll = self.ins.recvfrom(x)
+        pkt, sa_ll, ts = self._recv_raw(self.ins, x)
         if sa_ll[2] == socket.PACKET_OUTGOING:
             return None
         if sa_ll[3] in conf.l2types:
@@ -167,15 +266,17 @@ class L3RawSocket(SuperSocket):
             pkt = pkt.payload
 
         if pkt is not None:
-            from scapy.arch import get_last_packet_timestamp
-            pkt.time = get_last_packet_timestamp(self.ins)
+            if ts is None:
+                from scapy.arch import get_last_packet_timestamp
+                ts = get_last_packet_timestamp(self.ins)
+            pkt.time = ts
         return pkt
 
     def send(self, x):
         try:
             sx = raw(x)
             x.sent_time = time.time()
-            self.outs.sendto(sx, (x.dst, 0))
+            return self.outs.sendto(sx, (x.dst, 0))
         except socket.error as msg:
             log_runtime.error(msg)
 
@@ -190,6 +291,7 @@ class SimpleSocket(SuperSocket):
 
 class StreamSocket(SimpleSocket):
     desc = "transforms a stream socket into a layer 2"
+    nonblocking_socket = True
 
     def __init__(self, sock, basecls=None):
         if basecls is None:
@@ -201,7 +303,7 @@ class StreamSocket(SimpleSocket):
         pkt = self.ins.recv(x, socket.MSG_PEEK)
         x = len(pkt)
         if x == 0:
-            raise socket.error((100, "Underlying stream socket tore down"))
+            return None
         pkt = self.basecls(pkt)
         pad = pkt.getlayer(conf.padding_layer)
         if pad is not None and pad.underlayer is not None:
@@ -257,16 +359,16 @@ class L2ListenTcpdump(SuperSocket):
                  prog=None, *arg, **karg):
         self.outs = None
         args = ['-w', '-', '-s', '65535']
+        if iface is None and (WINDOWS or DARWIN):
+            iface = conf.iface
+        if WINDOWS:
+            try:
+                iface = iface.pcap_name
+            except AttributeError:
+                pass
+        self.iface = iface
         if iface is not None:
-            if WINDOWS:
-                try:
-                    args.extend(['-i', iface.pcap_name])
-                except AttributeError:
-                    args.extend(['-i', iface])
-            else:
-                args.extend(['-i', iface])
-        elif WINDOWS or DARWIN:
-            args.extend(['-i', conf.iface.pcap_name if WINDOWS else conf.iface])  # noqa: E501
+            args.extend(['-i', self.iface])
         if not promisc:
             args.append('-p')
         if not nofilter:
@@ -287,6 +389,12 @@ class L2ListenTcpdump(SuperSocket):
         SuperSocket.close(self)
         self.tcpdump_proc.kill()
 
+    @staticmethod
+    def select(sockets, remain=None):
+        if (WINDOWS or DARWIN):
+            return sockets, None
+        return SuperSocket.select(sockets, remain=remain)
+
 
 class TunTapInterface(SuperSocket):
     """A socket to act as the host's peer of a tun / tap interface.
@@ -296,7 +404,7 @@ class TunTapInterface(SuperSocket):
 
     def __init__(self, iface=None, mode_tun=None, *arg, **karg):
         self.iface = conf.iface if iface is None else iface
-        self.mode_tun = ("tun" in iface) if mode_tun is None else mode_tun
+        self.mode_tun = ("tun" in self.iface) if mode_tun is None else mode_tun
         self.closed = True
         self.open()
 
@@ -315,7 +423,8 @@ class TunTapInterface(SuperSocket):
             # IFF_TAP = 0x0002
             # IFF_NO_PI = 0x1000
             ioctl(self.ins, 0x400454ca, struct.pack(
-                "16sH", raw(self.iface), 0x0001 if self.mode_tun else 0x1002,
+                "16sH", bytes_encode(self.iface),
+                0x0001 if self.mode_tun else 0x1002,
             ))
         self.closed = False
 
@@ -337,8 +446,6 @@ conf.L2listen, conf.L2socket or conf.L3socket.
 
     def send(self, x):
         sx = raw(x)
-        if hasattr(x, "sent_time"):
-            x.sent_time = time.time()
         if self.mode_tun:
             try:
                 proto = conf.l3types[type(x)]
@@ -351,6 +458,10 @@ conf.L2listen, conf.L2socket or conf.L3socket.
                 proto = 0
             sx = struct.pack('!HH', 0, proto) + sx
         try:
-            os.write(self.outs.fileno(), sx)
+            try:
+                x.sent_time = time.time()
+            except AttributeError:
+                pass
+            return os.write(self.outs.fileno(), sx)
         except socket.error:
             log_runtime.error("%s send", self.__class__.__name__, exc_info=True)  # noqa: E501
