@@ -13,8 +13,8 @@ import time
 import struct
 import re
 import random
+import select
 import socket
-from select import select
 from collections import defaultdict
 
 from scapy.utils import checksum, do_graph, incremental_label, \
@@ -24,7 +24,7 @@ from scapy.data import ETH_P_IP, ETH_P_ALL, DLT_RAW, DLT_RAW_ALT, DLT_IPV4, \
     IP_PROTOS, TCP_SERVICES, UDP_SERVICES
 from scapy.layers.l2 import Ether, Dot3, getmacbyip, CookedLinux, GRE, SNAP, \
     Loopback
-from scapy.compat import raw, chb, orb
+from scapy.compat import raw, chb, orb, bytes_encode
 from scapy.config import conf
 from scapy.extlib import plt, MATPLOTLIB, MATPLOTLIB_INLINED, \
     MATPLOTLIB_DEFAULT_PLOT_KARGS
@@ -216,6 +216,28 @@ class IPOption_Traceroute(IPOption):
                    IPField("originator_ip", "0.0.0.0")]
 
 
+class IPOption_Timestamp(IPOption):
+    name = "IP Option Timestamp"
+    optclass = 2
+    option = 4
+    fields_desc = [_IPOption_HDR,
+                   ByteField("length", None),
+                   ByteField("pointer", 9),
+                   BitField("oflw", 0, 4),
+                   BitEnumField("flg", 1, 4,
+                                {0: "timestamp_only",
+                                 1: "timestamp_and_ip_addr",
+                                 3: "prespecified_ip_addr"}),
+                   ConditionalField(IPField("internet_address", "0.0.0.0"),
+                                    lambda pkt: pkt.flg != 0),
+                   IntField('timestamp', 0)]
+
+    def post_build(self, p, pay):
+        if self.length is None:
+            p = p[:1] + struct.pack("!B", len(p)) + p[2:]
+        return p + pay
+
+
 class IPOption_Address_Extension(IPOption):
     name = "IP Option Address Extension"
     copy_flag = 1
@@ -289,7 +311,10 @@ class RandTCPOptions(VolatileValue):
         # Pseudo-Random amount of options
         # Random ("NAME", fmt)
         rand_patterns = [
-            random.choice(list(TCPOptions[0].values())[1:])
+            random.choice(list(
+                (opt, fmt) for opt, fmt in six.itervalues(TCPOptions[0])
+                if opt != 'EOL'
+            ))
             for _ in range(self.size)
         ]
         rand_vals = []
@@ -331,13 +356,15 @@ class TCPOptionsField(StrField):
             onum = orb(x[0])
             if onum == 0:
                 opt.append(("EOL", None))
-                x = x[1:]
                 break
             if onum == 1:
                 opt.append(("NOP", None))
                 x = x[1:]
                 continue
-            olen = orb(x[1])
+            try:
+                olen = orb(x[1])
+            except IndexError:
+                olen = 0
             if olen < 2:
                 warning("Malformed TCP option (announced length is %i)" % olen)
                 olen = 2
@@ -378,7 +405,8 @@ class TCPOptionsField(StrField):
                     ofmt = TCPOptions[0][onum][1]
                     if onum == 5:  # SAck
                         ofmt += "%iI" % len(oval)
-                    if ofmt is not None and (not isinstance(oval, str) or "s" in ofmt):  # noqa: E501
+                    _test_isinstance = not isinstance(oval, (bytes, str))
+                    if ofmt is not None and (_test_isinstance or "s" in ofmt):
                         if not isinstance(oval, tuple):
                             oval = (oval,)
                         oval = struct.pack(ofmt, *oval)
@@ -393,7 +421,9 @@ class TCPOptionsField(StrField):
                 if not isinstance(oval, (bytes, str)):
                     warning("option [%i] is not bytes." % onum)
                     continue
-            opt += chb(onum) + chb(2 + len(oval)) + raw(oval)
+            if isinstance(oval, str):
+                oval = bytes_encode(oval)
+            opt += chb(onum) + chb(2 + len(oval)) + oval
         return opt + b"\x00" * (3 - ((len(opt) + 3) % 4))  # Padding
 
     def randval(self):
@@ -416,7 +446,7 @@ class ICMPTimeStampField(IntField):
         if isinstance(val, str):
             hmsms = self.re_hmsm.match(val)
             if hmsms:
-                h, _, m, _, s, _, ms = hmsms = hmsms.groups()
+                h, _, m, _, s, _, ms = hmsms.groups()
                 ms = int(((ms or "") + "000")[:3])
                 val = ((int(h) * 60 + int(m or 0)) * 60 + int(s or 0)) * 1000 + ms  # noqa: E501
             else:
@@ -543,7 +573,8 @@ class IP(Packet, IPTools):
 
     def fragment(self, fragsize=1480):
         """Fragment IP datagrams"""
-        fragsize = (fragsize + 7) // 8 * 8
+        lastfragsz = fragsize
+        fragsize -= fragsize % 8
         lst = []
         fnb = 0
         fl = self
@@ -553,7 +584,11 @@ class IP(Packet, IPTools):
 
         for p in fl:
             s = raw(p[fnb].payload)
-            nb = (len(s) + fragsize - 1) // fragsize
+            if len(s) <= lastfragsz:
+                lst.append(p)
+                continue
+
+            nb = (len(s) - lastfragsz + fragsize - 1) // fragsize + 1
             for i in range(nb):
                 q = p.copy()
                 del(q[fnb].payload)
@@ -561,8 +596,11 @@ class IP(Packet, IPTools):
                 del(q[fnb].len)
                 if i != nb - 1:
                     q[fnb].flags |= 1
+                    fragend = (i + 1) * fragsize
+                else:
+                    fragend = i * fragsize + lastfragsz
                 q[fnb].frag += i * fragsize // 8
-                r = conf.raw_layer(load=s[i * fragsize:(i + 1) * fragsize])
+                r = conf.raw_layer(load=s[i * fragsize:fragend])
                 r.overload_fields = p[fnb].payload.overload_fields.copy()
                 q.add_payload(r)
                 lst.append(q)
@@ -616,8 +654,10 @@ class TCP(Packet):
         p += pay
         dataofs = self.dataofs
         if dataofs is None:
-            dataofs = 5 + ((len(self.get_field("options").i2m(self, self.options)) + 3) // 4)  # noqa: E501
-            p = p[:12] + chb((dataofs << 4) | orb(p[12]) & 0x0f) + p[13:]
+            opt_len = len(self.get_field("options").i2m(self, self.options))
+            dataofs = 5 + ((opt_len + 3) // 4)
+            dataofs = (dataofs << 4) | orb(p[12]) & 0x0f
+            p = p[:12] + chb(dataofs & 0xff) + p[13:]
         if self.chksum is None:
             if isinstance(self.underlayer, IP):
                 ck = in4_chksum(socket.IPPROTO_TCP, self.underlayer, p)
@@ -848,13 +888,21 @@ class IPerror(IP):
     def answers(self, other):
         if not isinstance(other, IP):
             return 0
-        if not (((conf.checkIPsrc == 0) or (self.dst == other.dst)) and
-                (self.src == other.src) and
-                (((conf.checkIPID == 0) or
-                  (self.id == other.id) or
-                  (conf.checkIPID == 1 and self.id == socket.htons(other.id)))) and  # noqa: E501
-                (self.proto == other.proto)):
+
+        # Check if IP addresses match
+        test_IPsrc = not conf.checkIPsrc or self.src == other.src
+        test_IPdst = self.dst == other.dst
+
+        # Check if IP ids match
+        test_IPid = not conf.checkIPID or self.id == other.id
+        test_IPid |= conf.checkIPID and self.id == socket.htons(other.id)
+
+        # Check if IP protocols match
+        test_IPproto = self.proto == other.proto
+
+        if not (test_IPsrc and test_IPdst and test_IPid and test_IPproto):
             return 0
+
         return self.payload.answers(other.payload)
 
     def mysummary(self):
@@ -961,11 +1009,12 @@ conf.neighbor.register_l3(Dot3, IP, inet_register_l3)
 @conf.commands.register
 def fragment(pkt, fragsize=1480):
     """Fragment a big IP datagram"""
-    fragsize = (fragsize + 7) // 8 * 8
+    lastfragsz = fragsize
+    fragsize -= fragsize % 8
     lst = []
     for p in pkt:
         s = raw(p[IP].payload)
-        nb = (len(s) + fragsize - 1) // fragsize
+        nb = (len(s) - lastfragsz + fragsize - 1) // fragsize + 1
         for i in range(nb):
             q = p.copy()
             del(q[IP].payload)
@@ -973,8 +1022,11 @@ def fragment(pkt, fragsize=1480):
             del(q[IP].len)
             if i != nb - 1:
                 q[IP].flags |= 1
+                fragend = (i + 1) * fragsize
+            else:
+                fragend = i * fragsize + lastfragsz
             q[IP].frag += i * fragsize // 8
-            r = conf.raw_layer(load=s[i * fragsize:(i + 1) * fragsize])
+            r = conf.raw_layer(load=s[i * fragsize:fragend])
             r.overload_fields = p[IP].payload.overload_fields.copy()
             q.add_payload(r)
             lst.append(q)
@@ -1001,6 +1053,46 @@ overlap_fragsize: the fragment size of the overlapping packet"""
     return qfrag + fragment(p, fragsize)
 
 
+def _defrag_list(lst, defrag, missfrag):
+    """Internal usage only. Part of the _defrag_logic"""
+    p = lst[0]
+    lastp = lst[-1]
+    if p.frag > 0 or lastp.flags.MF:  # first or last fragment missing
+        missfrag.extend(lst)
+        return
+    p = p.copy()
+    if conf.padding_layer in p:
+        del(p[conf.padding_layer].underlayer.payload)
+    ip = p[IP]
+    if ip.len is None or ip.ihl is None:
+        clen = len(ip.payload)
+    else:
+        clen = ip.len - (ip.ihl << 2)
+    txt = conf.raw_layer()
+    for q in lst[1:]:
+        if clen != q.frag << 3:  # Wrong fragmentation offset
+            if clen > q.frag << 3:
+                warning("Fragment overlap (%i > %i) %r || %r ||  %r" % (clen, q.frag << 3, p, txt, q))  # noqa: E501
+            missfrag.extend(lst)
+            break
+        if q[IP].len is None or q[IP].ihl is None:
+            clen += len(q[IP].payload)
+        else:
+            clen += q[IP].len - (q[IP].ihl << 2)
+        if conf.padding_layer in q:
+            del(q[conf.padding_layer].underlayer.payload)
+        txt.add_payload(q[IP].payload.copy())
+        if q.time > p.time:
+            p.time = q.time
+    else:
+        ip.flags.MF = False
+        del(ip.chksum)
+        del(ip.len)
+        p = p / txt
+        p._defrag_pos = max(x._defrag_pos for x in lst)
+        defrag.append(p)
+
+
 def _defrag_logic(plist, complete=False):
     """Internal function used to defragment a list of packets.
     It contains the logic behind the defrag() and defragment() functions
@@ -1013,7 +1105,7 @@ def _defrag_logic(plist, complete=False):
         pos += 1
         if IP in p:
             ip = p[IP]
-            if ip.frag != 0 or ip.flags & 1:
+            if ip.frag != 0 or ip.flags.MF:
                 uniq = (ip.id, ip.src, ip.dst, ip.proto)
                 frags[uniq].append(p)
                 continue
@@ -1023,42 +1115,7 @@ def _defrag_logic(plist, complete=False):
     missfrag = []
     for lst in six.itervalues(frags):
         lst.sort(key=lambda x: x.frag)
-        p = lst[0]
-        lastp = lst[-1]
-        if p.frag > 0 or lastp.flags & 1 != 0:  # first or last fragment missing  # noqa: E501
-            missfrag.append(lst)
-            continue
-        p = p.copy()
-        if conf.padding_layer in p:
-            del(p[conf.padding_layer].underlayer.payload)
-        ip = p[IP]
-        if ip.len is None or ip.ihl is None:
-            clen = len(ip.payload)
-        else:
-            clen = ip.len - (ip.ihl << 2)
-        txt = conf.raw_layer()
-        for q in lst[1:]:
-            if clen != q.frag << 3:  # Wrong fragmentation offset
-                if clen > q.frag << 3:
-                    warning("Fragment overlap (%i > %i) %r || %r ||  %r" % (clen, q.frag << 3, p, txt, q))  # noqa: E501
-                missfrag.append(lst)
-                break
-            if q[IP].len is None or q[IP].ihl is None:
-                clen += len(q[IP].payload)
-            else:
-                clen += q[IP].len - (q[IP].ihl << 2)
-            if conf.padding_layer in q:
-                del(q[conf.padding_layer].underlayer.payload)
-            txt.add_payload(q[IP].payload.copy())
-            if q.time > p.time:
-                p.time = q.time
-        else:
-            ip.flags &= ~1  # !MF
-            del(ip.chksum)
-            del(ip.len)
-            p = p / txt
-            p._defrag_pos = max(x._defrag_pos for x in lst)
-            defrag.append(p)
+        _defrag_list(lst, defrag, missfrag)
     defrag2 = []
     for p in defrag:
         q = p.__class__(raw(p))
@@ -1149,7 +1206,7 @@ class TracerouteResult(SndRcvList):
                  "nloc"]
 
     def __init__(self, res=None, name="Traceroute", stats=None):
-        PacketList.__init__(self, res, name, stats)
+        SndRcvList.__init__(self, res, name, stats)
         self.graphdef = None
         self.graphASres = None
         self.padding = 0
@@ -1175,9 +1232,9 @@ class TracerouteResult(SndRcvList):
                 m = min(x for x, y in six.iteritems(k) if y[1])
             except ValueError:
                 continue
-            for l in list(k):  # use list(): k is modified in the loop
-                if l > m:
-                    del k[l]
+            for li in list(k):  # use list(): k is modified in the loop
+                if li > m:
+                    del k[li]
         return trace
 
     def trace3D(self, join=True):
@@ -1273,7 +1330,6 @@ Touch screen: pinch/extend to zoom, swipe or two-finger rotate."""
             )
         )
         vpython.scene.exit = True
-        start = vpython.box()
         rings = {}
         tr3d = {}
         for i in trace:
@@ -1623,8 +1679,17 @@ Touch screen: pinch/extend to zoom, swipe or two-finger rotate."""
 @conf.commands.register
 def traceroute(target, dport=80, minttl=1, maxttl=30, sport=RandShort(), l4=None, filter=None, timeout=2, verbose=None, **kargs):  # noqa: E501
     """Instant TCP traceroute
-traceroute(target, [maxttl=30,] [dport=80,] [sport=80,] [verbose=conf.verb]) -> None  # noqa: E501
-"""
+
+       :param target:  hostnames or IP addresses
+       :param dport:   TCP destination port (default is 80)
+       :param minttl:  minimum TTL (default is 1)
+       :param maxttl:  maximum TTL (default is 30)
+       :param sport:   TCP source port (default is random)
+       :param l4:      use a Scapy packet instead of TCP
+       :param filter:  BPF filter applied to received packets
+       :param timeout: time to wait for answers (default is 2s)
+       :param verbose: detailed output
+       :return: an TracerouteResult, and a list of unanswered packets"""
     if verbose is None:
         verbose = conf.verb
     if filter is None:
@@ -1651,9 +1716,10 @@ traceroute(target, [maxttl=30,] [dport=80,] [sport=80,] [verbose=conf.verb]) ->
 def traceroute_map(ips, **kargs):
     """Util function to call traceroute on multiple targets, then
     show the different paths on a map.
-    params:
-     - ips: a list of IPs on which traceroute will be called
-     - optional: kwargs, passed to traceroute"""
+
+    :param ips: a list of IPs on which traceroute will be called
+    :param kargs: (optional) kwargs, passed to traceroute
+    """
     kargs.setdefault("verbose", 0)
     return traceroute(ips)[0].world_trace()
 
@@ -1663,25 +1729,39 @@ def traceroute_map(ips, **kargs):
 
 
 class TCP_client(Automaton):
+    """
+    Creates a TCP Client Automaton.
+    This automaton will handle TCP 3-way handshake.
 
+    Usage: the easiest usage is to use it as a SuperSocket.
+        >>> a = TCP_client.tcplink(HTTP, "www.google.com", 80)
+        >>> a.send(HTTPRequest())
+        >>> a.recv()
+
+    :param ip: the ip to connect to
+    :param port:
+    """
     def parse_args(self, ip, port, *args, **kargs):
+        from scapy.sessions import TCPSession
         self.dst = str(Net(ip))
         self.dport = port
         self.sport = random.randrange(0, 2**16)
         self.l4 = IP(dst=ip) / TCP(sport=self.sport, dport=self.dport, flags=0,
                                    seq=random.randrange(0, 2**32))
         self.src = self.l4.src
-        self.swin = self.l4[TCP].window
-        self.dwin = 1
-        self.rcvbuf = b""
+        self.sack = self.l4[TCP].ack
+        self.rel_seq = None
+        self.rcvbuf = TCPSession(prn=self._transmit_packet, store=False)
         bpf = "host %s  and host %s and port %i and port %i" % (self.src,
                                                                 self.dst,
                                                                 self.sport,
                                                                 self.dport)
-
-#        bpf=None
         Automaton.parse_args(self, filter=bpf, **kargs)
 
+    def _transmit_packet(self, pkt):
+        """Transmits a packet from TCPSession to the SuperSocket"""
+        self.oi.tcp.send(raw(pkt[TCP].payload))
+
     def master_filter(self, pkt):
         return (IP in pkt and
                 pkt[IP].src == self.dst and
@@ -1690,7 +1770,7 @@ class TCP_client(Automaton):
                 pkt[TCP].sport == self.dport and
                 pkt[TCP].dport == self.sport and
                 self.l4[TCP].seq >= pkt[TCP].ack and  # XXX: seq/ack 2^32 wrap up  # noqa: E501
-                ((self.l4[TCP].ack == 0) or (self.l4[TCP].ack <= pkt[TCP].seq <= self.l4[TCP].ack + self.swin)))  # noqa: E501
+                ((self.l4[TCP].ack == 0) or (self.sack <= pkt[TCP].seq <= self.l4[TCP].ack + pkt[TCP].window)))  # noqa: E501
 
     @ATMT.state(initial=1)
     def START(self):
@@ -1724,7 +1804,7 @@ class TCP_client(Automaton):
 
     @ATMT.receive_condition(SYN_SENT)
     def synack_received(self, pkt):
-        if pkt[TCP].flags & 0x3f == 0x12:
+        if pkt[TCP].flags.SA:
             raise self.ESTABLISHED().action_parameters(pkt)
 
     @ATMT.action(synack_received)
@@ -1735,20 +1815,20 @@ class TCP_client(Automaton):
 
     @ATMT.receive_condition(ESTABLISHED)
     def incoming_data_received(self, pkt):
-        if not isinstance(pkt[TCP].payload, NoPayload) and not isinstance(pkt[TCP].payload, conf.padding_layer):  # noqa: E501
+        if not isinstance(pkt[TCP].payload, (NoPayload, conf.padding_layer)):
             raise self.ESTABLISHED().action_parameters(pkt)
 
     @ATMT.action(incoming_data_received)
     def receive_data(self, pkt):
         data = raw(pkt[TCP].payload)
         if data and self.l4[TCP].ack == pkt[TCP].seq:
+            self.sack = self.l4[TCP].ack
             self.l4[TCP].ack += len(data)
             self.l4[TCP].flags = "A"
+            # Answer with an Ack
             self.send(self.l4)
-            self.rcvbuf += data
-            if pkt[TCP].flags.P:
-                self.oi.tcp.send(self.rcvbuf)
-                self.rcvbuf = b""
+            # Process data - will be sent to the SuperSocket through this
+            self.rcvbuf.on_packet_received(pkt)
 
     @ATMT.ioevent(ESTABLISHED, name="tcp", as_supersocket="tcplink")
     def outgoing_data_received(self, fd):
@@ -1762,12 +1842,12 @@ class TCP_client(Automaton):
 
     @ATMT.receive_condition(ESTABLISHED)
     def reset_received(self, pkt):
-        if pkt[TCP].flags & 4 != 0:
+        if pkt[TCP].flags.R:
             raise self.CLOSED()
 
     @ATMT.receive_condition(ESTABLISHED)
     def fin_received(self, pkt):
-        if pkt[TCP].flags & 0x1 == 1:
+        if pkt[TCP].flags.F:
             raise self.LAST_ACK().action_parameters(pkt)
 
     @ATMT.action(fin_received)
@@ -1779,7 +1859,7 @@ class TCP_client(Automaton):
 
     @ATMT.receive_condition(LAST_ACK)
     def ack_of_fin_received(self, pkt):
-        if pkt[TCP].flags & 0x3f == 0x10:
+        if pkt[TCP].flags.A:
             raise self.CLOSED()
 
 
@@ -1830,20 +1910,21 @@ funcpres: a function used to summarize packets"""
 
 
 @conf.commands.register
-def fragleak(target, sport=123, dport=123, timeout=0.2, onlyasc=0):
+def fragleak(target, sport=123, dport=123, timeout=0.2, onlyasc=0, count=None):
     load = "XXXXYYYYYYYYYY"
-#    getmacbyip(target)
-#    pkt = IP(dst=target, id=RandShort(), options=b"\x22"*40)/UDP()/load
-    pkt = IP(dst=target, id=RandShort(), options=b"\x00" * 40, flags=1) / UDP(sport=sport, dport=sport) / load  # noqa: E501
+    pkt = IP(dst=target, id=RandShort(), options=b"\x00" * 40, flags=1)
+    pkt /= UDP(sport=sport, dport=sport) / load
     s = conf.L3socket()
     intr = 0
     found = {}
     try:
-        while True:
+        while count is None or count:
+            if count is not None and isinstance(count, int):
+                count -= 1
             try:
                 if not intr:
                     s.send(pkt)
-                sin, sout, serr = select([s], [], [], timeout)
+                sin = select.select([s], [], [], timeout)[0]
                 if not sin:
                     continue
                 ans = s.recv(1600)
@@ -1856,19 +1937,9 @@ def fragleak(target, sport=123, dport=123, timeout=0.2, onlyasc=0):
                 if ans.payload.payload.dst != target:
                     continue
                 if ans.src != target:
-                    print("leak from", ans.src, end=' ')
-
-
-#                print repr(ans)
+                    print("leak from", ans.src)
                 if not ans.haslayer(conf.padding_layer):
                     continue
-
-
-#                print repr(ans.payload.payload.payload.payload)
-
-#                if not isinstance(ans.payload.payload.payload.payload, conf.raw_layer):  # noqa: E501
-#                    continue
-#                leak = ans.payload.payload.payload.payload.load[len(load):]
                 leak = ans.getlayer(conf.padding_layer).load
                 if leak not in found:
                     found[leak] = None
@@ -1882,11 +1953,16 @@ def fragleak(target, sport=123, dport=123, timeout=0.2, onlyasc=0):
 
 
 @conf.commands.register
-def fragleak2(target, timeout=0.4, onlyasc=0):
+def fragleak2(target, timeout=0.4, onlyasc=0, count=None):
     found = {}
     try:
-        while True:
-            p = sr1(IP(dst=target, options=b"\x00" * 40, proto=200) / "XXXXYYYYYYYYYYYY", timeout=timeout, verbose=0)  # noqa: E501
+        while count is None or count:
+            if count is not None and isinstance(count, int):
+                count -= 1
+
+            pkt = IP(dst=target, options=b"\x00" * 40, proto=200)
+            pkt /= "XXXXYYYYYYYYYYYY"
+            p = sr1(pkt, timeout=timeout, verbose=0)
             if not p:
                 continue
             if conf.padding_layer in p:
