@@ -21,14 +21,17 @@ from scapy.compat import chb, orb
 from scapy.config import conf
 from scapy import consts
 from scapy.data import ARPHDR_ETHER, ARPHDR_LOOPBACK, ARPHDR_METRICOM, \
-    DLT_LINUX_IRDA, DLT_LINUX_SLL, DLT_LOOP, DLT_NULL, ETHER_ANY, \
-    ETHER_BROADCAST, ETHER_TYPES, ETH_P_ARP, ETH_P_MACSEC
-from scapy.error import warning
+    DLT_ETHERNET_MPACKET, DLT_LINUX_IRDA, DLT_LINUX_SLL, DLT_LOOP, \
+    DLT_NULL, ETHER_ANY, ETHER_BROADCAST, ETHER_TYPES, ETH_P_ARP, \
+    ETH_P_MACSEC
+from scapy.error import warning, ScapyNoDstMacException
 from scapy.fields import BCDFloatField, BitField, ByteField, \
-    ConditionalField, FieldLenField, IntEnumField, IntField, IP6Field, \
-    IPField, LenField, MACField, MultipleTypeField, ShortEnumField, \
-    ShortField, SourceIP6Field, SourceIPField, StrFixedLenField, StrLenField, \
-    X3BytesField, XByteField, XIntField, XShortEnumField, XShortField
+    ConditionalField, FieldLenField, FCSField, \
+    IntEnumField, IntField, IP6Field, IPField, \
+    LenField, MACField, MultipleTypeField, \
+    ShortEnumField, ShortField, SourceIP6Field, SourceIPField, \
+    StrFixedLenField, StrLenField, X3BytesField, XByteField, XIntField, \
+    XShortEnumField, XShortField
 from scapy.modules.six import viewitems
 from scapy.packet import bind_layers, Packet
 from scapy.plist import PacketList, SndRcvList
@@ -76,7 +79,7 @@ def getmacbyip(ip, chainCC=0):
     if (tmp[0] & 0xf0) == 0xe0:  # mcast @
         return "01:00:5e:%.2x:%.2x:%.2x" % (tmp[1] & 0x7f, tmp[2], tmp[3])
     iff, _, gw = conf.route.route(ip)
-    if ((iff == consts.LOOPBACK_INTERFACE) or (ip == conf.route.get_if_bcast(iff))):  # noqa: E501
+    if (iff == conf.loopback_name) or (ip in conf.route.get_if_bcast(iff)):
         return "ff:ff:ff:ff:ff:ff"
     if gw != "0.0.0.0":
         ip = gw
@@ -93,7 +96,8 @@ def getmacbyip(ip, chainCC=0):
                    verbose=0,
                    chainCC=chainCC,
                    nofilter=1)
-    except Exception:
+    except Exception as ex:
+        warning("getmacbyip failed on %s" % ex)
         return None
     if res is not None:
         mac = res.payload.hwsrc
@@ -115,8 +119,11 @@ class DestMACField(MACField):
             except socket.error:
                 pass
             if x is None:
-                x = "ff:ff:ff:ff:ff:ff"
-                warning("Mac address to reach destination not found. Using broadcast.")  # noqa: E501
+                if conf.raise_no_dst_mac:
+                    raise ScapyNoDstMacException()
+                else:
+                    x = "ff:ff:ff:ff:ff:ff"
+                    warning("Mac address to reach destination not found. Using broadcast.")  # noqa: E501
         return MACField.i2h(self, pkt, x)
 
     def i2m(self, pkt, x):
@@ -148,15 +155,10 @@ class SourceMACField(MACField):
         return MACField.i2m(self, pkt, self.i2h(pkt, x))
 
 
-class ARPSourceMACField(SourceMACField):
-    def __init__(self, name):
-        super(ARPSourceMACField, self).__init__(name)
-
-
 # Layers
 
-ETHER_TYPES['802_AD'] = 0x88a8
-ETHER_TYPES['802_1AE'] = ETH_P_MACSEC
+ETHER_TYPES[0x88a8] = '802_AD'
+ETHER_TYPES[ETH_P_MACSEC] = '802_1AE'
 
 
 class Ether(Packet):
@@ -189,7 +191,7 @@ class Ether(Packet):
 class Dot3(Packet):
     name = "802.3"
     fields_desc = [DestMACField("dst"),
-                   MACField("src", ETHER_ANY),
+                   SourceMACField("src"),
                    LenField("len", None, "H")]
 
     def extract_padding(self, s):
@@ -242,6 +244,13 @@ class CookedLinux(Packet):
                    XShortEnumField("proto", 0x800, ETHER_TYPES)]
 
 
+class MPacketPreamble(Packet):
+    # IEEE 802.3br Figure 99-3
+    name = "MPacket Preamble"
+    fields_desc = [StrFixedLenField("preamble", b"", length=8),
+                   FCSField("fcs", 0, fmt="!I")]
+
+
 class SNAP(Packet):
     name = "SNAP"
     fields_desc = [X3BytesField("OUI", 0x000000),
@@ -326,7 +335,7 @@ class ARP(Packet):
         }),
         MultipleTypeField(
             [
-                (ARPSourceMACField("hwsrc"),
+                (SourceMACField("hwsrc"),
                  (lambda pkt: pkt.hwtype == 1 and pkt.hwlen == 6,
                   lambda pkt, val: pkt.hwtype == 1 and (
                       pkt.hwlen == 6 or (pkt.hwlen is None and
@@ -400,10 +409,16 @@ class ARP(Packet):
         return self_psrc[:len(other_pdst)] == other_pdst[:len(self_psrc)]
 
     def route(self):
-        dst = self.pdst
+        fld, dst = self.getfield_and_val("pdst")
+        fld, dst = fld._find_fld_pkt_val(self, dst)
         if isinstance(dst, Gen):
             dst = next(iter(dst))
-        return conf.route.route(dst)
+        if isinstance(fld, IP6Field):
+            return conf.route6.route(dst)
+        elif isinstance(fld, IPField):
+            return conf.route.route(dst)
+        else:
+            return None, None, None
 
     def extract_padding(self, s):
         return "", s
@@ -434,6 +449,9 @@ class GRErouting(Packet):
 
 class GRE(Packet):
     name = "GRE"
+    deprecated_fields = {
+        "seqence_number": ("sequence_number", "2.4.4"),
+    }
     fields_desc = [BitField("chksum_present", 0, 1),
                    BitField("routing_present", 0, 1),
                    BitField("key_present", 0, 1),
@@ -446,7 +464,7 @@ class GRE(Packet):
                    ConditionalField(XShortField("chksum", None), lambda pkt:pkt.chksum_present == 1 or pkt.routing_present == 1),  # noqa: E501
                    ConditionalField(XShortField("offset", None), lambda pkt:pkt.chksum_present == 1 or pkt.routing_present == 1),  # noqa: E501
                    ConditionalField(XIntField("key", None), lambda pkt:pkt.key_present == 1),  # noqa: E501
-                   ConditionalField(XIntField("seqence_number", None), lambda pkt:pkt.seqnum_present == 1),  # noqa: E501
+                   ConditionalField(XIntField("sequence_number", None), lambda pkt:pkt.seqnum_present == 1),  # noqa: E501
                    ]
 
     @classmethod
@@ -471,6 +489,9 @@ class GRE_PPTP(GRE):
     """
 
     name = "GRE PPTP"
+    deprecated_fields = {
+        "seqence_number": ("sequence_number", "2.4.4"),
+    }
     fields_desc = [BitField("chksum_present", 0, 1),
                    BitField("routing_present", 0, 1),
                    BitField("key_present", 1, 1),
@@ -483,7 +504,7 @@ class GRE_PPTP(GRE):
                    XShortEnumField("proto", 0x880b, ETHER_TYPES),
                    ShortField("payload_len", None),
                    ShortField("call_id", None),
-                   ConditionalField(XIntField("seqence_number", None), lambda pkt: pkt.seqnum_present == 1),  # noqa: E501
+                   ConditionalField(XIntField("sequence_number", None), lambda pkt: pkt.seqnum_present == 1),  # noqa: E501
                    ConditionalField(XIntField("ack_number", None), lambda pkt: pkt.acknum_present == 1)]  # noqa: E501
 
     def post_build(self, p, pay):
@@ -515,7 +536,7 @@ LOOPBACK_TYPES = {0x2: "IPv4",
 
 
 class Loopback(Packet):
-    """*BSD loopback layer"""
+    r"""\*BSD loopback layer"""
 
     name = "Loopback"
     if consts.OPENBSD:
@@ -529,7 +550,7 @@ class Dot1AD(Dot1Q):
     name = '802_1AD'
 
 
-bind_layers(Dot3, LLC,)
+bind_layers(Dot3, LLC)
 bind_layers(Ether, LLC, type=122)
 bind_layers(Ether, LLC, type=34928)
 bind_layers(Ether, Dot1Q, type=33024)
@@ -544,6 +565,7 @@ bind_layers(CookedLinux, Dot1Q, proto=33024)
 bind_layers(CookedLinux, Dot1AD, type=0x88a8)
 bind_layers(CookedLinux, Ether, proto=1)
 bind_layers(CookedLinux, ARP, proto=2054)
+bind_layers(MPacketPreamble, Ether)
 bind_layers(GRE, LLC, proto=122)
 bind_layers(GRE, Dot1Q, proto=33024)
 bind_layers(GRE, Dot1AD, type=0x88a8)
@@ -551,7 +573,7 @@ bind_layers(GRE, Ether, proto=0x6558)
 bind_layers(GRE, ARP, proto=2054)
 bind_layers(GRE, GRErouting, {"routing_present": 1})
 bind_layers(GRErouting, conf.raw_layer, {"address_family": 0, "SRE_len": 0})
-bind_layers(GRErouting, GRErouting, {})
+bind_layers(GRErouting, GRErouting)
 bind_layers(LLC, STP, dsap=66, ssap=66, ctrl=3)
 bind_layers(LLC, SNAP, dsap=170, ssap=170, ctrl=3)
 bind_layers(SNAP, Dot1Q, code=33024)
@@ -565,6 +587,7 @@ conf.l2types.register_num2layer(ARPHDR_METRICOM, Ether)
 conf.l2types.register_num2layer(ARPHDR_LOOPBACK, Ether)
 conf.l2types.register_layer2num(ARPHDR_ETHER, Dot3)
 conf.l2types.register(DLT_LINUX_SLL, CookedLinux)
+conf.l2types.register(DLT_ETHERNET_MPACKET, MPacketPreamble)
 conf.l2types.register_num2layer(DLT_LINUX_IRDA, CookedLinux)
 conf.l2types.register(DLT_LOOP, Loopback)
 conf.l2types.register_num2layer(DLT_NULL, Loopback)
@@ -597,8 +620,21 @@ class ARPingResult(SndRcvList):
         SndRcvList.__init__(self, res, name, stats)
 
     def show(self):
+        """
+        Print the list of discovered MAC addresses.
+        """
+
+        data = list()
+        padding = 0
+
         for s, r in self.res:
-            print(r.sprintf("%19s,Ether.src% %ARP.psrc%"))
+            manuf = conf.manufdb._get_short_manuf(r.src)
+            manuf = "unknown" if manuf == r.src else manuf
+            padding = max(padding, len(manuf))
+            data.append((r[Ether].src, manuf, r[ARP].psrc))
+
+        for src, manuf, psrc in data:
+            print("  %-17s %-*s %s" % (src, padding, manuf, psrc))
 
 
 @conf.commands.register
@@ -615,7 +651,7 @@ Set cache=True if you want arping to modify internal ARP-Cache"""
     if cache and ans is not None:
         for pair in ans:
             conf.netcache.arp_cache[pair[1].psrc] = (pair[1].hwsrc, time.time())  # noqa: E501
-    if verbose:
+    if ans is not None and verbose:
         ans.show()
     return ans, unans
 
@@ -646,16 +682,24 @@ class ARP_am(AnsweringMachine):
 
     example:
     To respond to an ARP request for 192.168.100 replying on the
-    ingress interface;
+    ingress interface::
+
       farpd(IP_addr='192.168.1.100',ARP_addr='00:01:02:03:04:05')
-    To respond on a different interface add the interface parameter
+
+    To respond on a different interface add the interface parameter::
+
       farpd(IP_addr='192.168.1.100',ARP_addr='00:01:02:03:04:05',iface='eth0')
-    To respond on ANY arp request on an interface with mac address ARP_addr
+
+    To respond on ANY arp request on an interface with mac address ARP_addr::
+
       farpd(ARP_addr='00:01:02:03:04:05',iface='eth1')
-    To respond on ANY arp request with my mac addr on the given interface
+
+    To respond on ANY arp request with my mac addr on the given interface::
+
       farpd(iface='eth1')
 
-    Optional Args
+    Optional Args::
+
      inter=<n>   Interval in seconds between ARP replies being sent
 
     """
@@ -687,7 +731,6 @@ class ARP_am(AnsweringMachine):
                 ARP_addr = get_if_hwaddr(iff)
             except Exception:
                 ARP_addr = "00:00:00:00:00:00"
-                pass
         else:
             ARP_addr = self.ARP_addr
         resp = Ether(dst=ether.src,
